#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "Triangulate.h"


bool IsInTriangle(TTPoint cp, TTPoint p1, TTPoint p2, TTPoint p3)
{
#if 0
	// check cross point (if not crossing, point is out of triangle)
	float f,t,s;
	f = (cp.x-p1.x)*(p3.y-p2.y) - (cp.y-p1.y)*(p3.x-p2.x);
	if(f == 0)
		return false;
	t =((p1.y-p2.y)*(p3.x-p2.x) - (p1.x-p2.x)*(p3.y-p2.y)) / f;  //p1->cp
	s =((p1.y-p2.y)*(cp.x-p1.x) - (p1.x-p2.x)*(cp.y-p1.y)) / f;  //p2->p3
	if(t<1 || s<0 || 1<s)
		return false;
	return true;
#else
	if((p2.x-p1.x) * (cp.y-p2.y) - (p2.y-p1.y) * (cp.x-p2.x) < 0)
		return false;
	if((p3.x-p2.x) * (cp.y-p3.y) - (p3.y-p2.y) * (cp.x-p3.x) < 0)
		return false;
	if((p1.x-p3.x) * (cp.y-p1.y) - (p1.y-p3.y) * (cp.x-p1.x) < 0)
		return false;
	return true; // inside or over line
#endif
}

float GetCrossingAngle(TTPoint &v1, TTPoint &v2)
{
	float d1,d2,c;

	d1 = v1.GetSize();
	d2 = v2.GetSize();
	if(d1==0 || d2==0)
		return 0.0f;

	c = GetInnerProduct(v1, v2) / d1 / d2;
	if(c >  1.0f) c =  1.0f;
	if(c < -1.0f) c = -1.0f;

	return (float)acos(c);
}

Indices::Indices() {
	index = NULL;
	count = 0;
}

Indices::Indices(int i) {
	index = (int *)malloc(sizeof(int) * i);
	for(int j=0; j<i; j++)
		index[j] = j;
	count = i;
}

Indices::~Indices()
{
	if(index != NULL)
		free(index);
}

int Indices::Add(int i) {
	index = (int *)realloc(index, sizeof(int) * (count+1));
	index[count] = i;
	return count++;
}

int Indices::Insert(int i, int newindex) {
	index = (int *)realloc(index, sizeof(int) * (count+1));
	for(int j=count-1; j>=i; j--)
		index[j+1] = index[j];
	index[i] = newindex;
	count++;
	return i;
}

Indices *Indices::Clone(void)
{
	Indices *newindex = new Indices;
	for(int i=0; i<count; i++)
		newindex->Add(index[i]);
	return newindex;
}


IndicesGroup::IndicesGroup() {
	index = NULL;
	count = 0;
};

IndicesGroup::~IndicesGroup() {
	for(int i=0; i<count; i++) {
		if(index[i] != NULL)
			delete index[i];
	}
	if(index != NULL)
		free(index);
}

int IndicesGroup::Add(Indices *newindex)
{
	index = (Indices **)realloc(index, sizeof(Indices*) * (count+1));
	index[count++] = newindex;
	return count;
}

Indices *IndicesGroup::RemoveTop(void)
{
	return index[--count];
}


void TTPolygon::Triangulate(IndicesGroup *group)
{
	int i;
	TTPoint p0,p1,p2,v0,v1;

		while(group->count > 0)
		{
			Indices *index = group->RemoveTop();
			int icount = index->count;
			if(icount == 3) {
				triangle->Add(index);
				continue;
			}

			float minangle = 2.0f * PI;
			int minindex = 0;

			for(i=0; i<icount; i++)
			{
				p0 = pts[index->index[(i+icount-1)%icount]];
				p1 = pts[index->index[i]];
				p2 = pts[index->index[(i+1)%icount]];
				v0 = p1 - p0;
				v1 = p2 - p1;
				float angle = GetCrossingAngle(p0-p1, v1);
				if(v0.x * v1.y - v0.y * v1.x < 0)
					angle = 2.0f*PI - angle;

				if(minangle > angle) {
					minangle = angle;
					minindex = i;
				}
			}

			int mvi[3],inner,i1;
			mvi[0] = index->index[(minindex+icount-1) % icount];
			mvi[1] = index->index[minindex];
			mvi[2] = index->index[(minindex+1) % icount];
			p0 = pts[mvi[0]];
			p1 = pts[mvi[1]];
			p2 = pts[mvi[2]];
			v0 = p1 - p0;
			v1 = p2 - p1;

			int invi = -1;
			float mindis = 0;
			for(inner=0; inner<icount; inner++)
			{
				i1 = index->index[inner];
				if(i1 == mvi[0] || i1 == mvi[1] || i1 == mvi[2])
					continue;

				if(IsInTriangle(pts[i1], p0, p1, p2))
				{
					TTPoint p = pts[i1] - p1;
					float dis = p.GetSize(); 
					if(invi < 0 || mindis > dis){
						invi = inner;
						mindis = dis;
					}
				}
			}

			Indices *newindex;
			if(invi >= 0) {
				newindex = new Indices;
				for(i=minindex; ; ) {
					newindex->Add(index->index[i]);
					if(i == invi)
						break;
					if(++i >= icount)
						i = 0;
				}
				group->Add(newindex);
				newindex = new Indices;
				for(i=invi; ; ) {
					newindex->Add(index->index[i]);
					if(i == minindex)
						break;
					if(++i >= icount)
						i = 0;
				}
				group->Add(newindex);
			} else {
				newindex = new Indices;
				for(i=0; i<3; i++)
					newindex->Add(mvi[i]);
				triangle->Add(newindex);
				newindex = new Indices;
				for(i=(minindex+1) % icount; ; ) {
					newindex->Add(index->index[i]);
					if(i == (minindex+icount-1) % icount)
						break;
					if(++i >= icount)
						i = 0;
				}
				group->Add(newindex);
			}

			delete index;
		}
	}


TTPolygon::TTPolygon()
{
	pts = NULL;
	ptscount = 0;
	indices = NULL;
	triangle = NULL;
}

TTPolygon::~TTPolygon()
{
	if(pts != NULL)
		free(pts);
	if(indices != NULL)
		delete indices;
	if(triangle != NULL)
		delete triangle;
}

int TTPolygon::Add(TTPoint p)
{
	pts = (TTPoint *)realloc(pts, sizeof(TTPoint) * (ptscount+1));
	pts[ptscount] = p;
	return ptscount++;
}

int TTPolygon::Insert(int index, TTPoint p)
{
	pts = (TTPoint *)realloc(pts, sizeof(TTPoint) * (ptscount+1));
	for(int i=ptscount-1; i>=index; i--)
		pts[i+1] = pts[i];
	pts[index] = p;
	ptscount++;
	return index;
}

int TTPolygon::AddIndex(int i)
{
	if(indices == NULL)
		indices = new Indices;
	return indices->Add(i);
}

void TTPolygon::InitIndex(void)
{
	if(indices != NULL)
		delete indices;

	indices = new Indices(ptscount);
}

int TTPolygon::Triangulate(void)
{
	if(triangle != NULL)
		delete triangle;
	triangle = new IndicesGroup;

	if(indices == NULL)
		indices = new Indices(ptscount);
		
	IndicesGroup *group = new IndicesGroup;
	group->Add(indices->Clone());
	Triangulate(group);
	delete group;

	return triangle->count;
}

void TTPolygon::Invert(void)
{
	int i,j;
	for(i=0,j=ptscount-1; i<j; i++,j--)
	{
		TTPoint tmp = pts[i];
		pts[i] = pts[j];
		pts[j] = tmp;
	}
}

bool TTPolygon::IsClockwise(void)
{
	float dp0x,dp0y,dp1x,dp1y;
	float d0,d1,t;
	int i0,i1,i2;
	float angle=0;

	if(ptscount < 3)
		return false;

	for(i0=0; i0<ptscount; i0++)
	{
		i1 = i0+1<ptscount ? i0+1 : 0;
		i2 = i0+2<ptscount ? i0+2 : i0+2-ptscount;

		dp0x = pts[i1].x - pts[i0].x;
		dp0y = pts[i1].y - pts[i0].y;
		d0 = (float)sqrt(dp0x*dp0x + dp0y*dp0y);
		dp1x = pts[i2].x - pts[i1].x;
		dp1y = pts[i2].y - pts[i1].y;
		d1 = (float)sqrt(dp1x*dp1x + dp1y*dp1y);
		if(d0>0 && d1>0){
			t = (dp0x*dp1x + dp0y*dp1y) / (d0*d1);				
			t = (t<1.f) ? (float)acos(t) : 0.f;
			if(dp0x*dp1y - dp1x*dp0y > 0)
				angle += t;
			else
				angle -= t;
		}
	}

	return (angle > 0);
}

